import os
import sys
import argparse
import logging
import signal

from lal.gpstime import tconvert

from . import __version__
from . import config
from .plugins import followups
from .event import LocklossEvent, find_events
from . import condor

##################################################

def analyze_event(event, plugins=None):
    """Execute all follow-up plugins for event

    """
    event.lock()

    if plugins:
        log_mode = 'a'
    else:
        log_mode = 'w'

    # log analysis to event log
    logger = logging.getLogger()
    handler = logging.FileHandler(event.gen_path('log'), mode=log_mode)
    handler.setFormatter(logging.Formatter(config.LOG_FMT))
    logger.addHandler(handler)

    event._set_version(__version__)
    event._set_status(None)

    logging.info("analysis version: {}".format(event.analysis_version))
    logging.info("event path: {}".format(event.path))
    logging.info("event url: {}".format(event.url))

    complete = True
    for name, func in followups.items():
        if plugins and name not in plugins:
            continue
        logging.info("executing followup: {}({})".format(
            name, event.gps))
        try:
            func(event)
        except:
            complete = False
            logging.exception("exception in {}:".format(name))

    if complete:
        logging.info("analysis complete")
        event._set_status(0)
    else:
        logging.warning("INCOMPLETE ANALYSIS")
        event._set_status(1)

    logger.removeHandler(handler)
    event.release()

    return complete


def analyze_condor(event):
    if event.analyzing:
        logging.warning("event analysis already in progress, aborting")
        return
    condor_dir = event.gen_path('.condor')
    try:
        os.makedirs(condor_dir)
    except:
        pass
    sub = condor.CondorSubmit(
        condor_dir,
        'analyze',
        [str(event.gps)],
        local=False,
    )
    sub.write()
    sub.submit()

##################################################

def _parser_add_arguments(parser):
    egroup = parser.add_mutually_exclusive_group(required=True)
    egroup.add_argument('event', type=float, nargs='?',
                        help="event ID / GPS second")
    egroup.add_argument('--condor', type=float, nargs=2, metavar='GPS',
                        help="condor analyze all events within GPS range")
    parser.add_argument('--rerun', action='store_true',
                        help="condor re-analyze events")
    parser.add_argument('--plugin', '-p', action='append',
                        help="execute only specified plugin (multiple ok, not available for condor runs)")


def main(args=None):
    """Analyze event(s)

    By default all analysis follow-ups for the specified event will
    exeuted.  If the --condor option is given with GPS start and end
    times all events within the specified time span will be analyzed.

    """
    if not args:
        parser = argparse.ArgumentParser()
        _parser_add_arguments(parser)
        args = parser.parse_args()

    if args.condor:
        logging.info("finding events...")
        after, before = tuple(int(t) for t in args.condor)
        events = [e for e in find_events(after=after, before=before)]
        to_analyze = [e for e in events
                      if not e.analyzing and (not e.analyzed or args.rerun)]
        logging.info("found {} events, {} to analyze.".format(len(events), len(to_analyze)))
        if not to_analyze:
            return

        def condor_args_gen():
            for e in to_analyze:
                yield [('event', e)]

        dag = condor.CondorDAG(
            config.CONDOR_ANALYZE_DIR,
            'analyze',
            condor_args_gen,
        )
        dag.write()
        dag.submit()

    else:
        event = LocklossEvent(args.event)

        if event.analyzing:
            raise SystemExit("Event analysis in progress: {}".format(os.path.join(event.path, 'log')))

        # FIXME: option to condor analyze single event?
        #analyze_condor(event)
        if not analyze_event(event, args.plugin):
            sys.exit(1)


# direct execution of this module intended for condor jobs
if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    logging.basicConfig(
        level='DEBUG',
        format=config.LOG_FMT,
    )
    parser = argparse.ArgumentParser()
    parser.add_argument('event', type=int,
                        help="event ID / GPS second")
    args = parser.parse_args()
    event = LocklossEvent(args.event)
    if not analyze_event(event):
        sys.exit(1)
