import numpy as np
import matplotlib.pyplot as plt
from matplotlib import rcParamsDefault

from gwpy.segments import Segment

from .. import config
from ..event import LocklossEvent
from .. import data

##############################################


def plot_lsc_asc(event):
    gps = int(event.gps)
    refined_gps = event.refined_gps
    if refined_gps:
        diff = refined_gps-gps
    else:
        diff = 0
    plot_windows = config.REFINE_PLOT_WINDOWS
    lsc_asc = config.LSC_ASC_CHANNELS

    # set plot metaparams
    plt.rcParams.update(rcParamsDefault)
    plt.rcParams['font.size'] = 20.0
    plt.rcParams['axes.titlesize'] = 20.0
    plt.rcParams['xtick.labelsize'] = 20.0
    plt.rcParams['ytick.labelsize'] = 20.0
    plt.rcParams['legend.fontsize'] = 20.0
    plt.rcParams['agg.path.chunksize'] = 1000

    for chan_group, channels in lsc_asc.items():
        for window_type, window in plot_windows.items():
            segment = Segment(window).shift(gps)
            bufs = data.fetch(channels, segment)
            for idx, buf in enumerate(bufs):
                color = 'royalblue'
                name = buf.channel
                srate = buf.sample_rate
                t = np.arange(window[0], window[1], 1/srate)
                fig, ax = plt.subplots(1, figsize=(12,8))
                ax.set_xlabel('Time [s] since refined lockloss at {}'.format(gps), labelpad=10)
                ax.set_ylabel('Counts')
                ax.set_xlim(window)
                ymin, ymax = calc_ylims(refined_gps, t, buf.data, diff)
                ax.set_ylim(ymin, ymax)
                ax.grid()
                ax.plot(t, buf.data, color = color, label = name, linewidth=2.5)
                ax.legend()

                ax.axvline(
                    diff,
                    color = 'black',
                    linestyle = '--',
                    lw = 3,
                )

                if event.refined_gps == event.gps:
                    annotation = 'Unrefined GPS'
                else:
                    annotation = 'Refined GPS'
                ax.annotate(
                    annotation,
                    xy=(diff, ax.get_ylim()[0]),
                    xycoords='data',
                    xytext=(0, 2),
                    textcoords='offset points',
                    horizontalalignment='center',
                    verticalalignment='bottom',
                    bbox=dict(boxstyle="round", fc="w", ec="black", alpha=0.95),
                )

                fig.tight_layout()
                # ax.set_title('Timeseries for %s' % (name))

                #saves plot to lockloss directory
                outfile_plot = '{}_{}.png'.format(name, window_type)
                outpath_plot = event.gen_path(outfile_plot)
                fig.savefig(outpath_plot, bbox_inches='tight')
                fig.clf()


def calc_ylims(refined_gps, t, y, diff):
    """
    Calculate ylims for the section of SUS channels occuring before refined/
    unrefined lockloss time.
    """

    scaling_times = np.where(t < diff-0.2)[0]
    scaling_vals = y[scaling_times]
    ymin, ymax = min(scaling_vals), max(scaling_vals)

    if ymax > 0:
        ymax = 1.1*ymax
    else:
        ymax = 0.9*ymax
    if ymin < 0:
        ymin = 1.1*ymin
    else:
        ymin = 0.9*ymin

    if ymin == ymax:
        ymin, ymax = -0.2, 0.2

    return ymin, ymax


def is_null_channel(data):

    null = None
    if all(data == 0):
        null = True
    else:
        null = False

    return null

###########################################

def _parser_add_arguments(parser):
    parser.add_argument('event', type=int, nargs='?',
                        help="event ID")


def main(args=None):
    """
    Create plots of LSC/ASC channels around a lockloss.
    """
    if not args:
        parser = argparse.ArgumentParser()
        _parser_add_arguments(parser)
        args = parser.parse_args()

    event = LocklossEvent(args.event)
    plot_lsc_asc(event)


if __name__ == '__main__':
    main()
