import os
import glob

import bottle
from datetime import datetime, timedelta

from .. import config
from ..event import LocklossEvent, find_events
from . import templates, utils

# NOTE: shouldn't be needed, but url paths include this from apache
WEB_SCRIPT = config.WEB_ROOT + '/index.cgi'

##################################################

app = bottle.Bottle()


def event_list(query):
    # query info
    yield '<h3>query: {}</h3>'.format(
        ', '.join(['{}={}'.format(k, v) for k, v in query.items()]))
    yield '</h3>'

    # event table header
    yield bottle.template(templates.EVENT_TABLE_HEADER)

    # event table row
    for i, event in enumerate(find_events(**query)):
        if i >= query['show']:
            break
        previous_state = event.previous_state
        if previous_state:
            previous = previous_state[0]
            duration = str(timedelta(seconds=int(previous_state[2] - previous_state[1])))
        else:
            previous = event.transition_index[0]
            duration = ''
        yield bottle.template(
            templates.EVENT_ROW,
            web_script=WEB_SCRIPT,
            index=i,
            gps=event.gps,
            utc=event.utc,
            previous=previous,
            duration=duration,
            tags=utils.tag_buttons(event.list_tags()),
            status_button=utils.analysis_status_button(event),
        )

    # footers
    yield bottle.template(templates.EVENT_TABLE_FOOTER)


def event_summary(gps):
    # event summary
    try:
        event = LocklossEvent(gps)
    except OSError as e:
        yield '<h3>Unknown event: {}</h3>'.format(gps)
        return

    yield utils.analysis_status_button(event)
    yield '<h3>lockloss: <a href="{}">{}</a></h3>'.format(
        event.url,
        event.gps,
    )
    yield '<h5>{}</h5>'.format(event.utc)
    p, l = event.transition_index
    yield '<h6>state transition: <tt>{node}</tt>  {prev} -> {lockloss}</h6>'.format(
        node=config.GRD_NODE,
        prev=p,
        lockloss=l,
    )
    if event.refined_gps:
        yield '<h6>refined GPS: {}</h6>'.format(event.refined_gps)
    yield '<h6>tags: {}</h6>'.format(utils.tag_buttons(event.list_tags()))

    ### indicator plot
    yield '<div class="container">'
    yield '<div class="row">'
    for plot_url in utils.event_plot_urls(event, 'indicators'):
        yield bottle.template(templates.IMAGE, url=plot_url, size=6)
    yield '</div>'
    yield '</div>'

    ### saturation plots
    yield '<hr /><div class="container">'
    yield '<h3>Saturation Plots</h3><br />'
    yield '<div class="row">'
    csv_path = event.gen_path('saturations.csv')
    if os.path.exists(csv_path):
        ### link plots
        for plot_url in utils.event_plot_urls(event, 'saturations'):
            yield bottle.template(templates.IMAGE, url=plot_url, size=6)

        yield bottle.template(templates.SATURATION_TABLE_HEADER)

        ### saturation table
        with open(csv_path, 'r') as f:
            for line in f:
                sat_channel, sat_time = line.split()
                yield bottle.template(
                    templates.SATURATION_TABLE_ROW,
                    channel=sat_channel,
                    time=sat_time,
                )

        yield bottle.template(templates.SATURATION_TABLE_FOOTER)
    else:
        yield '<p>No saturating suspension channels before refined lockloss time.</p>'
    yield '</div>'
    yield '</div>'

    ### LPY plots
    yield '<hr /><div class="container">'
    yield '<h3>Length-Pitch-Yaw Plots</h3><br />'
    yield '<p>DAC counts for first suspension stage to saturate</p>'
    yield '<div class="row">'
    if not event.analyzed:
        yield '<p>analysis underway...</p>'
    for plot_url in utils.event_plot_urls(event, 'lpy'):
        yield bottle.template(templates.IMAGE, url=plot_url, size=6)
    if not os.path.exists(csv_path):
        yield '<p>LPY plots not created due to lack of saturating suspension channels.</p>'
    yield '</div>'
    yield '</div>'

    ### ADC overflow plots
    yield '<hr /><div class="container">'
    yield '<h3>ADC Overflow Plots</h3><br />'
    yield '<h5>Overview</h5>'

    # overview plots
    yield '<div class="row">'
    if not event.analyzed:
        yield '<p>analysis underway...</p>'
    for plot_url in utils.event_plot_urls(event, 'adc_overflow_overview'):
        yield bottle.template(templates.IMAGE, url=plot_url, size=6)
    yield '</div>'

    # specific ADC plots
    for ii, subsystem in enumerate(['ASC', 'LSC', 'OMC']):
        adc_plots = sorted(glob.glob(event.gen_path('adc_overflow_{}_*.png'.format(subsystem))))

        yield bottle.template(templates.ADC_OVERFLOW_HEADER, idx=ii, subsystem=subsystem)
        yield '<div class="row">'
        for adc_plot in adc_plots:
            adc_name = os.path.basename(adc_plot)
            sat_url = event.gen_url(adc_name)
            yield bottle.template(templates.IMAGE, url=sat_url, size=6)
        yield '</div>'
        yield bottle.template(templates.ADC_OVERFLOW_FOOTER)
    yield '</div>'

    ### glitch plots
    yield '<hr /><div class="container">'
    yield '<h3>Glitch Event Plots</h3><br />'
    yield '<div class="panel-group">'
    yield '<div class="panel panel-default">'
    yield '<div class="panel-heading">'
    yield '<h5 class="panel-title"><a data-toggle="collapse" href="#glitch_plots">Click to show plots</a></h5>'
    yield '</div>'
    yield '<div id="glitch_plots" class="panel-collapse collapse">'
    yield '<div class="panel-body">'
    yield '<div class="container">'

    for ii, subsystem in enumerate(config.GLITCH_CHANNELS.keys()):
        plot_name = 'snr_eventmap_{}'.format(subsystem)
        plot_urls = utils.event_plot_urls(event, plot_name)
        yield '<div class="row">'
        if not plot_urls:
            yield '<p>No glitch plots available for {} subsystem</p>'.format(subsystem)
        else:
            for plot_url in plot_urls:
                yield bottle.template(templates.IMAGE, url=plot_url, size=6)
        yield '</div>'
    yield '</div></div></div></div>'

    ### LSC/ASC plots
    yield '<hr /><div class="container">'
    yield '<h3>LSC/ASC Plots</h3><br />'
    yield '<p> Lists of relevant LSC/ASC channels </p>'

    for chan_group, chan_list in config.LSC_ASC_CHANNELS.items():
        yield '<h5>{}</h5>'.format(chan_group)
        yield '<div class="row">'
        for ii, channel in enumerate(chan_list):
            # for zoom_level in ['WIDE', 'ZOOM']:
            for zoom_level in ['WIDE']:
                plot_name = '{}_{}.png'.format(channel, zoom_level)
                plot_path = event.gen_path(plot_name)
                lsc_asc_name = os.path.basename(plot_path)
                sat_url = event.gen_url(lsc_asc_name)
                yield bottle.template(templates.THUMBNAIL, url=sat_url, size=3)
        yield '</div>'
        yield '<br>'

##########

@app.route("/")
def index():
    if bottle.request.query.format and bottle.request.query.format == 'json':
        query = utils.query_to_str(dict(bottle.request.query))
        bottle.redirect('{}/api/tag/{}?{}'.format(WEB_SCRIPT, tag, query))
        return

    elif 'event' in bottle.request.query:
        gps = bottle.request.query.get('event')
        # bottle.redirect('{}/event/{}'.format(WEB_SCRIPT, gps))
        body_generator = event_summary(gps)
        meta_refresh = ''

    else:
        query = utils.query_to_dict(dict(bottle.request.query))
        body_generator = event_list(query)
        meta_refresh = '<meta http-equiv="refresh" content="30">'

    FULL_HEAD = \
        templates.HEADER + \
        templates.HEADER_INFO + \
        '<hr />' + \
        templates.EVENT_FORM + \
        templates.FILTER_FORM + \
        '<hr />'

    yield bottle.template(
        FULL_HEAD,
        IFO=config.IFO,
        web_script=WEB_SCRIPT,
        meta_refresh=meta_refresh,
        date=datetime.now(),
        online_status=utils.online_status(),
    )

    for o in body_generator:
        yield o

    yield bottle.template(templates.FOOTER)


@app.route("/event/<gps>")
def event_route(gps):
    bottle.redirect('{}?event={}'.format(WEB_SCRIPT, gps))


@app.route("/tag/<tag>")
def tag_route(tag='all'):
    query = utils.query_to_str(dict(bottle.request.query))
    bottle.redirect('{}?tag={}&{}'.format(WEB_SCRIPT, tag, query))


@app.route("/api/tag/<tag>")
def json_route(tag='all'):
    query = utils.query_to_dict(dict(bottle.request.query))
    events = []
    for i, event in enumerate(find_events(**query)):
        if i > query['show']:
            break
        events.append(event.to_dict())
    return {'events': events}

##################################################

if __name__ == '__main__':
    bottle.run(app, server='cgi', debug=True)
